/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.api.AlignViewportI;
import jalview.datamodel.SequenceI;
import jalview.renderer.AnnotationRenderer;
import jalview.renderer.seqfeatures.FeatureColourFinder;
import jalview.viewmodel.OverviewDimensions;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.image.BufferedImage;

import javax.swing.JComponent;


public class OverviewCanvas extends JComponent
{
  private static final Color TRANS_GREY = new Color(100, 100, 100, 25);

  // This is set true if the alignment view changes whilst
  // the overview is being calculated
  private volatile boolean restart = false;

  private volatile boolean updaterunning = false;

  private BufferedImage miniMe;

  private BufferedImage lastMiniMe = null;

  // Can set different properties in this seqCanvas than
  // main visible SeqCanvas
  private SequenceRenderer sr;

  private jalview.renderer.seqfeatures.FeatureRenderer fr;

  private final AnnotationRenderer renderer = new AnnotationRenderer();

  OverviewDimensions od;

  AlignViewportI av;

  public OverviewCanvas(OverviewDimensions overviewDims,
          AlignViewportI alignvp)
  {
    od = overviewDims;
    av = alignvp;

    sr = new SequenceRenderer(av);
    sr.renderGaps = false;
    sr.forOverview = true;
    fr = new jalview.renderer.seqfeatures.FeatureRenderer(av);
  }

  /*
   * Signals to drawing code that the associated alignment viewport
   * has changed and a redraw will be required
   */
  public boolean restartDraw()
  {
    synchronized (this)
    {
      if (updaterunning)
      {
        restart = true;
      }
      else
      {
        updaterunning = true;
      }
      return restart;
    }
  }

  public void draw(boolean showSequenceFeatures, boolean showAnnotation,
          FeatureRenderer transferRenderer)
  {
    miniMe = null;

    if (showSequenceFeatures)
    {
      fr.transferSettings(transferRenderer);
    }

    // why do we need to set preferred size again? was set in
    // updateOverviewImage
    setPreferredSize(new Dimension(od.getWidth(), od.getHeight()));

    miniMe = new BufferedImage(od.getWidth(), od.getHeight(),
            BufferedImage.TYPE_INT_RGB);

    Graphics mg = miniMe.getGraphics();
    mg.setColor(Color.orange);
    mg.fillRect(0, 0, od.getWidth(), miniMe.getHeight());

    // calculate sampleCol and sampleRow
    // alignment width is max number of residues/bases
    // alignment height is number of sequences
    int alwidth = av.getAlignment().getWidth();
    int alheight = av.getAlignment().getAbsoluteHeight();

    // sampleCol or sampleRow is the width/height allocated to each residue
    // in particular, sometimes we may need more than one row/col of the
    // BufferedImage allocated
    // sampleCol is how much of a residue to assign to each pixel
    // sampleRow is how many sequences to assign to each pixel
    float sampleCol = alwidth / (float) od.getWidth();
    float sampleRow = alheight / (float) od.getSequencesHeight();

    buildImage(sampleRow, sampleCol);

    if (showAnnotation)
    {
      renderer.updateFromAlignViewport(av);
      for (int col = 0; col < od.getWidth() && !restart; col++)
      {
        mg.translate(col, od.getSequencesHeight());
        renderer.drawGraph(mg, av.getAlignmentConservationAnnotation(),
                av.getAlignmentConservationAnnotation().annotations,
                (int) (sampleCol) + 1, od.getGraphHeight(),
                (int) (col * sampleCol), (int) (col * sampleCol) + 1);
        mg.translate(-col, -od.getSequencesHeight());

      }
    }
    System.gc();

    if (restart)
    {
      restart = false;
      draw(showSequenceFeatures, showAnnotation, transferRenderer);
    }
    else
    {
      updaterunning = false;
      lastMiniMe = miniMe;
    }
  }

  @Override
  public void paintComponent(Graphics g)
  {
    if (restart)
    {
      if (lastMiniMe == null)
      {
        g.setColor(Color.white);
        g.fillRect(0, 0, getWidth(), getHeight());
      }
      else
      {
        g.drawImage(lastMiniMe, 0, 0, getWidth(), getHeight(), this);
      }
      g.setColor(TRANS_GREY);
      g.fillRect(0, 0, getWidth(), getHeight());
    }
    else if (lastMiniMe != null)
    {
      g.drawImage(lastMiniMe, 0, 0, this);
      if (lastMiniMe != miniMe)
      {
        g.setColor(TRANS_GREY);
        g.fillRect(0, 0, getWidth(), getHeight());
      }
    }

    g.setColor(Color.red);
    od.drawBox(g);
  }

  /*
   * Build the overview panel image
   */
  private void buildImage(float sampleRow, float sampleCol)
  {
    int lastcol = -1;
    int lastrow = -1;
    int rgbcolor = Color.white.getRGB();

    SequenceI seq = null;
    FeatureColourFinder finder = new FeatureColourFinder(fr);

    final boolean hasHiddenCols = av.hasHiddenColumns();
    boolean hiddenRow = false;
    // get hidden row and hidden column map once at beginning.
    // clone featureRenderer settings to avoid race conditions... if state is
    // updated just need to refresh again
    for (int row = 0; row < od.getSequencesHeight() && !restart; row++)
    {
      boolean doCopy = true;
      int currentrow = (int) (row * sampleRow);
      if (currentrow != lastrow)
      {
        doCopy = false;

        lastrow = currentrow;

        // get the sequence which would be at alignment index 'lastrow' if no
        // rows were hidden, and determine whether it is hidden or not
        hiddenRow = av.getAlignment().isHidden(lastrow);
        seq = av.getAlignment().getSequenceAtAbsoluteIndex(lastrow);
      }

      for (int col = 0; col < od.getWidth() && !restart; col++)
      {
        if (doCopy)
        {
          rgbcolor = miniMe.getRGB(col, row - 1);
        }
        else if ((int) (col * sampleCol) != lastcol
                || (int) (row * sampleRow) != lastrow)
        {
          lastcol = (int) (col * sampleCol);
          rgbcolor = getColumnColourFromSequence(seq, hiddenRow,
                  hasHiddenCols, lastcol, finder);
        }
        // else we just use the color we already have , so don't need to set it

        miniMe.setRGB(col, row, rgbcolor);
      }
    }
  }

  /*
   * Find the colour of a sequence at a specified column position
   */
  private int getColumnColourFromSequence(jalview.datamodel.SequenceI seq,
          boolean hiddenRow, boolean hasHiddenCols, int lastcol,
          FeatureColourFinder finder)
  {
    Color color = Color.white;

    if ((seq != null) && (seq.getLength() > lastcol))
    {
      color = sr.getResidueColour(seq, lastcol, finder);
    }

    if (hiddenRow
            || (hasHiddenCols && !av.getColumnSelection()
                    .isVisible(lastcol)))
    {
      color = color.darker().darker();
    }

    return color.getRGB();
  }

}
